
% June 19, 2009

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This file contains a number of utility functions that make using
% ISIS a little bit easier.  Variants of a few of these functions will
% show up as standard functions in ISIS v. 1.5.0 (due to be released
% Summer 2009).  The names and interfaces here are very close to what
% you will find in ISIS 1.5.  Among the functions expected to go into
% ISIS 1.5.0 will be:
%
% ignore_values     - called
% notice_values
% group             - called here group_sn_chan
% conf_loop
% save_par_limits
% add_unit
% unit_info
% is_energy_unit
% 
% Once ISIS v. 1.5.0 is released, stop using this file!  ISIS & S-lang
% let you overwrite most existing scripted functions with identically
% named functions that possibly do different things.  That's a useful
% feature in many ways, but of course can be dangerous.  
%
% For that reason, this file contains only those functions that will
% eventually make it into ISIS 1.5 in one form or another.  It's meant
% to be never used again, after one has installed ISIS 1.5.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define save_par_limits()
%!%+
%\function{save_par_limits}
%\synopsis{Save the results from a conf_loop error bar search to a file.}
%\usage{ save_par_limits(a, pmin, pmax [, file]);}
%\description
%
%   This saves the results from a \"conf_loop\" error bar search to a
%   parameter file by storing the lower/upper confidence limits in
%   the min/max columns.  This will *not* overwrite the parameters
%   being used during the current analysis session.
%
%   Inputs:
%
%      a   :  Array of parameter indices or names
%      pmin: Parameter lower confidence limit
%      pmax: Parameter upper confidence limit
%      file: Parameter save file name (default=\"conf_loop.save\")
%
%\seealso{conf_loop, conf, fconf,  vconf, conf_joint, conf_map_counts, conf_map_flux, plot_conf, save_conf, load_conf}
%!%-
{
   variable a, pmin, pmax, file, tol;
   switch(_NARGS)
   {
    case 3:
      (a,pmin,pmax) = ();
      file = "conf_loop.save";
   }
   {
    case 4:
      (a,pmin,pmax,file) = ();
   }
   {
      variable str = [""," save_par_limits(a,pmin,pmax [,file]);"," ",
      "   Save the results from a \"conf_loop\" error bar search to a",
      "   parameter file by storing the lower/upper confidence limits in",
      "   the min/max columns.  This will *not* overwrite the parameters",
      "   being used during the current analysis session.","",
      "   Inputs:","",
      "     a   :  Array of parameter indices or names",
      "     pmin: Parameter lower confidence limit",
      "     pmax: Parameter upper confidence limit",
      "     file: Parameter save file name (default=\"conf_loop.save\")",""];
      vmessage("%s",strjoin(str,"\n"));
      return;
   }
   if( (length(a) != length(pmin)) || (length(a) != length(pmax)) )
   {
      message("\n Inconsistent array lengths. \n");
      return;
   }
   if(typeof(file) != String_Type)
   {
      message("\n Must input a String_Type file name. \n");
      return;
   }  
   tol = qualifier("tol",1.e-3);
   variable ii=0, i, b, params;
   params = get_params(a);
   foreach(a)
   {
      i = ();
      b = get_par(i);
      set_par(i,b,0,pmin[ii],pmax[ii]);
      ii++;
   }  
   save_par(file);
   set_params(params);
   return;
} 


%%%%%%%%%%%%%%%%%%%%%%%%%

public define conf_loop()
%!%+
%\function{conf_loop}
%\synopsis{Search for errors, but refit if a lower chi^2 is found.}
%\usage{ (pmin,pmax) = conf_loop(a, [,level [,tol]]; qualifiers);}
%\qualifiers{
%\qualifier{nmax}{[100]: Number of new fits to try before quitting}
%\qualifier{prefix}{[\"conf_loop\"]: Root name for save files}
%\qualifier{save}{Set qualifier to write out files}
%\qualifier{tol}{[1.e-3]: pmin/pmax will be set to at least (1-tol)/(1+tol)*parameter}
%}
%\description
%
%  conf_loop() is a simple wrapper around the conf() function that
%  takes an array of parameter indices or names, and then serially
%  determines the confidence limits for each.  The order for the
%  confidence limit search follows that of the input array.  (See,
%  however, the description below of the refit procedure.)
%  Alternatively, one can input the string \"all\", and the code will
%  perform confidence limit searches on all free/untied parameters.
%  The code will return, without attempting any confidence limit
%  searches, if any of the input parameters are frozen or tied.
%
%  By default, the function searches for the 90% confidence limits for
%  one interesting parameter (i.e., Delta chi^2 ~ 2.71). 68% and 95%
%  confidence limits can instead be saved by changing the level
%  parameter.  As for conf, 0=68%, 1=90%, 2=95% confidence limits.
%
%  Using an input string ('prefix') as a root name, the function first
%  saves the current fit parameters to: prefix.par.  As intermediate
%  results are obtained, the confidence limits for a parameter are
%  stored in: prefix.conf.  The final confidence limits, along with the
%  final best fit parameters, are saved as a parameter file:
%  prefix.save.  The latter file has its min/max columns replaced with
%  the lower/upper confidence limits.  The current best fit parameters
%  from the analysis session, however, retain their initial input
%  min/max values.
%
%  In the case of no input file root name, the default is to write
%  results to conf_loop.par, conf_loop.conf, and conf_loop.save in the
%  current working directory.  Should these files not be able to be
%  opened, an attempt will be made to write them in the /tmp
%  directory.  If that fails, no output will be saved.  File output
%  can also be turned off by setting the \"nosave\" qualifier to a
%  nonzero value.
%
%  Should a new, lower chi^2 value be obtained during the confidence
%  limit search process, the model will be refit, the new best fit
%  will be stored in prefix.par, and the error bar search will restart
%  from the parameter whose confidence limit search led to the refit.
%  These new confidence limits will be written to prefix.conf, without
%  erasing the old history of confidence limits.  prefix.save, however,
%  will only be written once conf_loop finishes successfully. This
%  refit procedure will be applied up to nmax (default=100) times
%  before the confidence limit search is abandoned.
%
%  Upon completion, vectors containing the obtained upper and lower
%  confidence limits (in the order of the input parameter list, or in
%  index order if \"all\" was input) will be returned.  Any parameter
%  with an unsearched limit (since nmax refits was exceeded) will have
%  its corresponding pmin/pmax set to 0.
%
%  Examples:
%
%     (pmin,pmax) = conf_loop(["powerlaw(1).norm","powerlaw(1).PhoIndex"];save);
%
%  This searches for the 90% confidence limits for the normalization
%  and index of powerlaw(1), and saves the results to conf_loop.par,
%  conf_loop.conf, conf_loop.save.  pmin and pmax are Double_Type
%  arrays of length 2.
%
%     (pmin,pmax) = conf_loop("all",0;prefix="powerlaw_err",save);
%
%  This searches for 68% confidence limits for all the parameters, and
%  saves the results to powerlaw_err.par, powerlaw_err.conf,
%  powerlaw_err.save
%
%     (pmin,pmax) = conf_loop([1:10],2,0.3;prefix="error_bars",nmax=2,save);
%
%  This searches for 95% confidence limits for parameters 1 through 10
%  (presuming they are all unfrozen, untied, otherwise conf_loop will
%  exit returning NULLs).  Refits will only occur if the new chi^2
%  differs from the old chi^2 by 0.3 or more.  Refits will happen only
%  twice before abandoning the confidence limit search.  Results will
%  be saved to error_bars.par, error_bars.conf, error_bars.save.
%
%\seealso{save_par_limits, conf, fconf,  vconf, conf_joint, conf_map_counts, conf_map_flux, plot_conf, save_conf, load_conf}
%!%-
{
   variable a, nmax, level=NULL, toler=NULL, nosave, file, filec, filep, files, info;

   switch(_NARGS)
   {
    case 1:
      a = ();
   }
   {
    case 2:
      (a,level) = ();
   }
   {
    case 3:
      (a,level,toler) = ();
   }
   {
      variable str = ["",
      " (pmin,pmax) = conf_loop(a, [,level [,tolerance]] ;qualifiers);",
      " ",
      "   Search for errors, but refit if a lower chi^2 is found.",
      "",
      " Inputs:","",
      "   a        : Vector giving indices or names of parameters for error", 
      "              search, or the string \"all\" to search all free parameters",
      "   level    : 0, 1, 2 = 68%, 90%, 99% confidence level (default 90%)",
      "   tolerance: Minimum Delta chi^2 required for a refit (default 1.e-3).","",
      " Qualifiers:","",
      "   nmax  = Number of new fits to try before quitting (default=100)",
      "   prefix= Root name for intermediate results (prefix.par, prefix.conf),",
      "           and final results (prefix.save - see \"save_par_limits();\")",
      "           If file open fails, attempts will be made to write results",
      "           to /tmp/conf_loop.conf and /tmp/conf_loop.save instead",
      "   save  = Include to write out files, otherwise no writing occurs","",
      " Outputs:","",
      "   pmin,pmax: Vector with lower, upper parameter bounds for the errors",
      "              obtained from either a successful fit, or from quitting",
      "              after nmax new fits. (pmin/pmax are set to 0 if the",
      "              parameter limit has not been searched)",""];
      vmessage("%s",strjoin(str,"\n"));
      return;
   }

   if(level==NULL){ level = 1; }
   if(toler==NULL){ toler = 1.e-3; }

   nmax = qualifier("nmax",100);
   file = qualifier("prefix","conf_loop");
   nosave = 1;

   if(qualifier_exists("save")){ nosave=0; };

   if(level == NULL){level = 1; }
   if(toler == NULL){toler = 1.e-3; }

   filec = file+".conf";
   filep = file+".par";
   files = file+".save";

   save_par(filep);
   
   variable spar, ii,n, aindx=Integer_Type[0];

   a = [a];
   if(typeof(a[0])==String_Type)
   {
      if(strlow(a[0])=="all")
      {
        _for ii (1,get_num_pars,1)
        {
           spar = get_par_info(ii);
           if(spar.freeze==0 && spar.tie==NULL)
           {
              aindx = [aindx,ii];
           }
        }
        a = aindx;
      }
   }

   variable la = length(a);
   variable i=0,j=0,imax=0,pmin=0.,pmax=0.;
   variable pmin_final = Float_Type[length(a)];
   variable pmax_final = Float_Type[length(a)];
   variable fun, vpar, tpar, pindx;

   foreach ii (a)
   {
      if(typeof(ii) != String_Type && ii > get_num_pars)
      {
        () = printf("\n Parameter %4i is not a valid parameter \n",ii);
        return NULL,NULL;
      }
      spar = get_par_info(ii);
      if(spar==NULL)
      {
         () = printf("\n Parameter %s not valid \n",ii);
         return NULL, NULL;
      }
      if(spar.freeze==1 or typeof(spar.tie)==String_Type)
      {
         () = printf("\n Parameter %4i is frozen or tied \n",spar.index);
         return NULL, NULL;
      }
   }

   fun = get_fit_fun();

   ifnot(nosave)
   {
      variable fp = fopen(filec,"w");
      if(fp==NULL)
       {  
         % Simple attempt to salvage something if file open fails
         message("\n File save failed - writing to /tmp/conf_loop.conf \n");
	 filec = "/tmp/conf_loop.conf";
	 files = "/tmp/conf_loop.save";
         fp = fopen(filec,"w");
         if(fp==NULL)
         {
            % No writes if further attempt fails
            message("\n File save failed - no saves will occur \n");
            nosave=1;
         }
      }
      ifnot(nosave)
      {
         () = fprintf(fp,"\n "+fun+" \n \n");
         () = fprintf(fp," ITER   PAR#      P_VALUE        P_MIN        P_MAX  TIME \n");
         () = fclose(fp);
      }
   }

   while(i < la and imax <= nmax)
   {
      _for ii (0, la-1-j, 1)
      {
         while( pmin == pmax and imax <= nmax )
         {
            (pmin,pmax) = conf(a[ii],level,toler);
            if(pmin == pmax)
            {
               () = fit_counts(&info);

               ifnot(nosave)
               {
                  % Any time we refit, save the parameters, including chi^2
                  save_par(filep);
                  tpar = time();

                  fp = fopen(filec,"a");
                  () = fprintf(fp,"\n chi^2 =%11.4e,  DoF =%6i, time = "+tpar+" \n \n",
                       info.statistic,(info.num_bins-info.num_variable_params));
                  () = fclose(fp);
               }

               i = 0;
               j = 0;
               imax++;
            }
            else
            {
               ifnot(nosave)
               {
                  tpar = time();
                  vpar = get_par(a[ii]);
                  pindx = get_par_info(a[ii]).index;

                  fp = fopen(filec,"a");
                  () = fprintf(fp,"%5i   %4i  %11.4e  %11.4e  %11.4e  "+tpar+" \n",
                               imax,pindx,vpar,pmin,pmax);
                  () = fclose(fp);
               }
               i++;
            }
         }
         pmin_final[ii] = pmin;
         pmax_final[ii] = pmax;
         pmin = 0.;
         pmax = 0.;
      }
      j=i;
   }
   if(imax <= nmax && nosave==0)
   {
      save_par_limits(a,pmin_final,pmax_final,files);
   }
   return pmin_final, pmax_final;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
private variable Units = Assoc_Type[];

define add_unit (name, is_energy, scale)
{
   variable u = struct
     {
        name = name,
        is_energy = is_energy,
        scale = scale
        % Multiplying by 'scale' converts energy units to keV
        % and wavelength units to Angstrom.
     };

   Units[strlow(name)] = u;
}

add_unit ("Angstrom",  0, 1.0);
add_unit ("A",  0, 1.0);
add_unit ("nm", 0, 1.e1);
add_unit ("um", 0, 1.e4);
add_unit ("mm", 0, 1.e7);
add_unit ("cm", 0, 1.e8);
add_unit ("m",  0, 1.e10);

add_unit ("eV",  1, 1.e-3);
add_unit ("keV", 1, 1.0);
add_unit ("MeV", 1, 1.e3);
add_unit ("GeV", 1, 1.e6);
add_unit ("TeV", 1, 1.e9);

private variable hz_to_kev = Const_h / (Const_eV * 1.e3);
add_unit ("Hz",  1, hz_to_kev);
add_unit ("kHz", 1, hz_to_kev * 1.e3);
add_unit ("MHz", 1, hz_to_kev * 1.e6);
add_unit ("GHz", 1, hz_to_kev * 1.e9);

private variable Default_Unit = "keV";

%%%%%%%%%%%%%%%%%%%%

define unit_info (s)
{
   variable lo_s = strlow(s);
   if (assoc_key_exists (Units, lo_s))
     return Units[lo_s];
   throw ApplicationError, "unrecognized spectrum coordinate '$s'"$;
}

%%%%%%%%%%%%%%%%%%%%%%

define to_angstrom (x)
{
   variable units, u, xa;
   units = qualifier ("unit", Default_Unit);

   u = unit_info (units);
   xa = x * u.scale;
   if (u.is_energy) xa = _A(xa);

   return xa;
}

%%%%%%%%%%%%%%%%%%%%%%%%%

define is_energy_unit (s)
{
   return unit_info(s).is_energy;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% binary search a histogram grid
define bsearch_hist (x, lo, hi)
{
   variable n = length(lo);
   if (n == 0)
     return -1;

   if (isnan(x))
     return -1;

   if (x < lo[0] || hi[n-1] < x)
     return -1;

   variable n0=0, n1=n, n2;

   while (n1 > n0 + 1)
     {
        n2 = (n0 + n1) / 2;
        if (x < hi[n2])
          {
             if (lo[n2] <= x)
               return n2;
             n1 = n2;
          }
        else n0 = n2;
     }

   return n0;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define init_data_struct_II (k)
{
   rebin_data (k, 0);

   variable s = struct {value, back, scaled_back, bin_lo, bin_hi};

   s.bin_lo = get_data_counts(k).bin_lo;
   s.bin_hi = get_data_counts(k).bin_hi;
   s.value = get_data_counts (k).value;
   s.back = get_back (k);

   variable scale = 1.0;
   if (s.back != NULL)
     {
        scale = [((get_data_exposure (k) * get_data_backscale(k))
                  /get_back_exposure (k) * get_back_backscale(k))];
        scale[where(isnan(scale))] = 1.0;
        if (length(scale) == 1) scale = scale[0];
     }
   else s.back = Double_Type[length(s.value)];

   s.scaled_back = scale * s.back;

   return s;
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define indices_for_ybounds (s, ybounds)
{
   variable k,
     ny = length(ybounds),
     ibounds = Integer_Type[ny];

   _for k (0, ny-1, 1)
     {
        ibounds[k] = bsearch_hist (ybounds[k], s.bin_lo, s.bin_hi);
     }

   print(ibounds);
   return ibounds[where (ibounds >= 0)];
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define force_array_length (type, n, q)
{
   variable value = type[n];

   if (q != NULL)
     {
        switch (length(q))
          {case 1:  value[*] = q;}
          {(length(q)< n):
             value[[0:length(q)-1]] = q;
             value[[length(q):n-1]] = q[-1];
          }
          {(length(q)>=n):  
            value = q[[0:n-1]];
          }
     }

   return value;

}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

public define group(datasets)
{
  variable d, data;

  variable datasets_use=[qualifier("sn_data",datasets)];

  datasets = [datasets];
  data = array_map(Struct_Type, &init_data_struct_II, datasets_use);

  variable 
    num_bins = length(data[0].value),
    rebin_flags = Integer_Type[num_bins];

  variable units = qualifier("unit", Default_Unit),
           bounds = qualifier("bounds",[0]);

  variable swap = is_energy_unit(units); 

  variable length_bounds = length(bounds), min_sn, min_chan;

  % Default S/N threshold corresponds to +/- 5% uncertainty
  min_sn = force_array_length(Double_Type, length_bounds,
                             qualifier("min_sn",1/sqrt(0.05)));

  min_chan = force_array_length(Integer_Type, length_bounds,
                               qualifier("min_chan",1));

  bounds = to_angstrom(bounds;;__qualifiers);

  if(swap)
  {
      min_sn = reverse(min_sn);
      min_chan = reverse(min_chan);
  }

  variable bound_comp, min_chan_comp, min_sn_comp, icomp;

  % Start from low end of Angstrom-type unit 
  variable lp_lw=0, lp_hi=num_bins-1, lp_inc=1; 
  % or from low end of keV-type unit
  if(swap)
  {
     lp_lw=num_bins-1; lp_hi=0; lp_inc=-1;
  }

  variable i, _sign = 1, bin = NULL;

  _for i (lp_lw, lp_hi, lp_inc)
  {
     rebin_flags[i] = _sign;

     if (bin == NULL)
      bin = struct {ssum=0, bsum=0, bsum_scl=0, num_chan=0};

     foreach d (data)
     {
        bin.ssum += d.value[i];
        bin.bsum += d.back[i];
        bin.bsum_scl += d.scaled_back[i];  
     }

     bin.num_chan++;

     % Figure out what minimum S/N and number of channels we are comparing to
     if(swap)
     {
        bound_comp = data[0].bin_hi[i];
        icomp = [where(bounds >= bound_comp),length_bounds-1];
        icomp = min(icomp);
     }
     else
     {
        bound_comp = data[0].bin_lo[i];
        icomp = [where(bounds <= bound_comp),0];
        icomp = max(icomp);
     }

     min_sn_comp = min_sn[icomp];
     min_chan_comp = min_chan[icomp];     

     if( ( swap==0 && (bound_comp < bounds[0])  ) || % Below lower bounds?...
         ( swap==1 && (bound_comp > bounds[-1]) ) || % ... then don't group
         ( (bin.num_chan >= min_chan_comp) &&              % Channels criteria
           ( ( min_sn_comp <= 0) ||                        % S/N not important
             ( min_sn_comp <= (bin.ssum-bin.bsum)/
                             sqrt(bin.ssum+bin.bsum_scl) ) % S/N important
           ) 
         ) 
       )
     {
        _sign *= -1;
        bin = NULL;
     }
  }

  foreach i (datasets)
  {
     rebin_data(i, rebin_flags);
  }
}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

private define mask_intervals ()
{
   variable nargs = _NARGS;
   _stk_roll(-nargs);
   variable s = ();
   nargs--;

   if (nargs mod 2 != 0)
     {
        _pop_n(nargs);
        throw ApplicationError;
     }

   variable len = length(s.bin_lo),
     mask = Int_Type[len];

   variable num_pairs = nargs / 2;
   while (num_pairs > 0)
     {
        variable lo, hi;
        (lo, hi) = ();

        variable xa = to_angstrom ([lo,hi];; __qualifiers);
        (lo, hi) = (xa[0], xa[1]);

        variable i = where (lo <= s.bin_lo and s.bin_hi <= hi);
        mask[i] = 1;

        num_pairs--;
     }

   return mask;
}

private define mask_values (datasets)
{
   variable d = get_data_counts(datasets[0]),
     len = length(d.value),
     veto = Int_Type[len];

   variable
     min_sum = qualifier ("min_sum", NULL),
     min_val = qualifier ("min_val", NULL);

   if (min_sum == NULL && min_val == NULL)
     return not veto;

   variable i, s = Double_Type[len];

   foreach i (datasets)
     {
        variable vi = get_data_counts(i).value;
        s += vi;
        if (min_val != NULL)
          {
             veto[where (vi < min_val)] = 1;
          }
     }

   if (min_sum != NULL)
     {
        veto[where (s < min_sum)] = 1;
     }

   return not veto;
}

private define mismatched_grids (datasets)
{
   if (length(datasets) == 1)
     return 0;

   variable i, s = get_data_counts (datasets[0]);
   foreach i (datasets[[1:]])
     {
        variable ss = get_data_counts (i);
        if (any (ss.bin_lo != s.bin_lo or ss.bin_hi != s.bin_hi))
          return 1;
     }
   return 0;
}

private define perform_select ()
{
   variable nargs = _NARGS;

   variable select_method = ();
   nargs--;

   _stk_roll (-nargs);
   variable datasets = ();
   nargs--;

   variable args = __pop_args (nargs);

   if (mismatched_grids (datasets))
     throw ApplicationError, "*** Error:  mismatched dataset grids";

   variable value_mask, interval_mask;
   value_mask = mask_values (datasets ;; __qualifiers);

   variable grid = get_data_counts (datasets[0]);
   interval_mask = mask_intervals (grid, __push_args(args) ;; __qualifiers);

   variable d;
   foreach d (datasets)
   {
      (@select_method)(d, where (value_mask and interval_mask));
   }
}

define ignore_values ()
{
   variable args = __pop_args (_NARGS);
   perform_select (__push_args (args), &ignore_list ;; __qualifiers);
}

define notice_values ()
{
   variable args = __pop_args (_NARGS);
   ignore(args[0].value);
   perform_select (__push_args (args), &notice_list ;; __qualifiers);
}

